import json

import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator20 import validate_spec


@pytest.fixture
def swagger_spec(petstore_contents):
    return json.loads(petstore_contents)


def test_failure_on_duplicate_api_parameters(swagger_spec):
    param1 = {"name": "foo", "in": "query", "type": "string"}
    param2 = {"name": "foo", "in": "query", "type": "integer"}
    swagger_spec["paths"]["/pet"]["parameters"] = [param1, param2]
    with pytest.raises(SwaggerValidationError) as exc_info:
        validate_spec(swagger_spec)
    assert "Duplicate param found with (name, in): {}".format(("foo", "query")) in str(
        exc_info.value
    )


def test_failure_on_duplicate_operation_parameters(swagger_spec):
    param1 = {"name": "foo", "in": "query", "type": "string"}
    param2 = {"name": "foo", "in": "query", "type": "integer"}
    swagger_spec["paths"]["/pet"]["post"]["parameters"].extend([param1, param2])
    with pytest.raises(SwaggerValidationError) as exc_info:
        validate_spec(swagger_spec)
    assert "Duplicate param found with (name, in): {}".format(("foo", "query")) in str(
        exc_info.value
    )


def test_failure_on_unresolvable_path_parameter(swagger_spec):
    swagger_spec["paths"]["/pet/{petId}"]["get"]["parameters"] = []

    with pytest.raises(SwaggerValidationError) as exc_info:
        validate_spec(swagger_spec)
    assert "Path parameter 'petId' used is not documented on '/pet/{petId}'" in str(
        exc_info.value
    )


def test_failure_on_path_parameter_used_but_not_defined(swagger_spec):
    swagger_spec["paths"]["/user/{username}"]["get"]["parameters"][0]["name"] = "_"
    with pytest.raises(SwaggerValidationError) as exc_info:
        validate_spec(swagger_spec)
    assert (
        "Path parameter 'username' used is not documented on '/user/{username}'"
        in str(exc_info.value)
    )


def test_failure_on_unresolvable_ref_of_props_required_list(swagger_spec):
    swagger_spec["definitions"]["Pet"]["required"].append("bla")
    with pytest.raises(
        SwaggerValidationError,
        match=r".*In definition of .*, required list has properties not defined: .*",
    ) as exc_info:
        validate_spec(swagger_spec)
    assert str(["bla"]) in str(exc_info.value)


def test_failure_on_unresolvable_model_reference_from_model(swagger_spec):
    swagger_spec["definitions"]["Pet"]["properties"]["category"]["$ref"] = "_"
    with pytest.raises(SwaggerValidationError):
        validate_spec(swagger_spec)


def test_failure_on_unresolvable_model_reference_from_param(swagger_spec):
    param = swagger_spec["paths"]["/pet"]["post"]["parameters"][0]
    param["schema"]["$ref"] = "#/definitions/bla"
    with pytest.raises(SwaggerValidationError):
        validate_spec(swagger_spec)


def test_failure_on_unresolvable_model_reference_from_resp(swagger_spec):
    resp = swagger_spec["paths"]["/pet/findByStatus"]["get"]["responses"]
    resp["200"]["schema"]["items"]["$ref"] = "#/definitions/bla"
    with pytest.raises(SwaggerValidationError):
        validate_spec(swagger_spec)


# TODO: Add warning validations for unused models, path parameter & responses
# TODO: Add validations for cyclic model definitions



import json

import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator20 import validate_json


def test_success():
    with open("./tests/data/v2.0/petstore.json") as f:
        petstore_spec = json.load(f)
    validate_json(petstore_spec, "schemas/v2.0/schema.json")


def test_failure():
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_json({}, "schemas/v2.0/schema.json")
    assert "'swagger' is a required property" in str(excinfo.value)



from __future__ import annotations

import logging
from types import ModuleType
from typing import Any

from swagger_spec_validator import validator12
from swagger_spec_validator import validator20
from swagger_spec_validator.common import read_url
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.common import wrap_exception


log = logging.getLogger(__name__)


def get_validator(spec_json: dict[str, Any], origin: str = "unknown") -> ModuleType:
    """
    :param spec_json: Dict representation of the json API spec
    :param origin: filename or url of the spec - only use for error messages
    :return: module responsible for validation based on Swagger version in the
        spec
    """
    swagger12_version = spec_json.get("swaggerVersion")
    swagger20_version = spec_json.get("swagger")

    if swagger12_version and swagger20_version:
        raise SwaggerValidationError(
            "You've got conflicting keys for the Swagger version in your spec. "
            "Expected `swaggerVersion` or `swagger`, but not both."
        )
    elif swagger12_version and swagger12_version == "1.2":
        # we don't care about versions prior to 1.2
        return validator12
    elif swagger20_version and swagger20_version == "2.0":
        return validator20
    elif swagger12_version is None and swagger20_version is None:
        raise SwaggerValidationError(
            "Swagger spec {} missing version. Expected "
            "`swaggerVersion` or `swagger`".format(origin)
        )
    else:
        raise SwaggerValidationError(
            "Swagger version {} not supported.".format(
                swagger12_version or swagger20_version
            )
        )


@wrap_exception
def validate_spec_url(spec_url: str) -> None:
    """Validates a Swagger spec given its URL.

    :param spec_url:
        For Swagger 1.2, this is the URL to the resource listing in api-docs.
        For Swagger 2.0, this is the URL to swagger.json in api-docs.
        If given as ``file://``, this must be an absolute url for cross-refs
        to work correctly.
    """
    spec_json = read_url(spec_url)
    validator = get_validator(spec_json, spec_url)
    validator.validate_spec(spec_json, spec_url)



from unittest.mock import Mock

import pytest
from jsonschema.exceptions import RefResolutionError
from jsonschema.validators import RefResolver

from swagger_spec_validator.validator20 import deref


def test_none():
    assert deref(None, Mock(spec=RefResolver)) is None


def test_not_dict():
    assert deref(1, Mock(spec=RefResolver)) == 1


def test_not_ref():
    input = {"type": "object"}
    assert deref(input, Mock(spec=RefResolver)) == input


def test_ref():
    ref_dict = {"$ref": "#/definitions/Foo"}
    definitions = {"definitions": {"Foo": "bar"}}
    assert deref(ref_dict, RefResolver("", definitions)) == "bar"


def test_ref_not_found():
    ref_dict = {"$ref": "#/definitions/Foo"}
    definitions = {}
    with pytest.raises(RefResolutionError) as excinfo:
        deref(ref_dict, RefResolver("", definitions))
    assert "Unresolvable JSON pointer" in str(excinfo.value)



import json
from unittest import mock

import pytest

from swagger_spec_validator.common import get_uri_from_file_path
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator12 import validate_spec_url
from tests import TESTS_BASE_PATH
from tests.conftest import is_urlopen_error


RESOURCE_LISTING_FILE = TESTS_BASE_PATH + "/data/v1.2/foo/swagger_api.json"
API_DECLARATION_FILE = TESTS_BASE_PATH + "/data/v1.2/foo/foo.json"


def read_contents(file_name):
    with open(file_name) as f:
        return json.load(f)


def make_mock_responses(file_names):
    return [read_contents(file_name) for file_name in file_names]


def test_http_success():
    mock_responses = make_mock_responses([RESOURCE_LISTING_FILE, API_DECLARATION_FILE])

    with mock.patch(
        "swagger_spec_validator.validator12.read_url",
        side_effect=mock_responses,
    ) as mock_read_url:
        validate_spec_url("http://localhost/api-docs")

        mock_read_url.assert_has_calls(
            [
                mock.call("http://localhost/api-docs"),
                mock.call("http://localhost/api-docs/foo"),
            ]
        )


def test_file_uri_success():
    mock_string = "swagger_spec_validator.validator12.validate_api_declaration"
    with mock.patch(mock_string) as mock_api:
        validate_spec_url(get_uri_from_file_path(RESOURCE_LISTING_FILE))

        expected = read_contents(API_DECLARATION_FILE)
        mock_api.assert_called_once_with(expected)


def test_raise_SwaggerValidationError_on_urlopen_error():
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec_url("http://foo")
    assert is_urlopen_error(excinfo.value)



import os

import pytest


def is_urlopen_error(exception):
    return any(
        urlopen_error_str in str(exception)
        for urlopen_error_str in {
            "<urlopen error [Errno -2] Name or service not known>",
            "<urlopen error [Errno -3] Temporary failure in name resolution>",
            "<urlopen error [Errno 8] nodename nor servname provided, or not known",
            "<urlopen error [Errno -5] No address associated with hostname>",
            "<urlopen error [Errno 11001] getaddrinfo failed>",
        }
    )


@pytest.fixture(autouse=True)
def change_dir():
    """
    Change the base directory to git root directory such that tests does not need to
    generate the absolute path for the data files.
    A relative path from git top-level directory will work as well
    """
    current_directory = os.path.abspath(os.curdir)
    try:
        os.chdir(os.path.abspath(os.path.dirname(os.path.dirname(__file__))))
        yield
    finally:
        os.chdir(current_directory)



import json

import pytest
from jsonschema.validators import RefResolver

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator20 import validate_spec
from tests.validator20.conftest import get_spec_json_and_url


@pytest.fixture
def minimal_swagger_dict():
    """Return minimal dict that respresents a swagger spec - useful as a base
    template.
    """
    return {
        "swagger": "2.0",
        "info": {
            "title": "Test",
            "version": "1.0",
        },
        "paths": {},
        "definitions": {},
    }


def test_success(petstore_dict):
    assert isinstance(validate_spec(petstore_dict), RefResolver)


def test_definitons_not_present_success(minimal_swagger_dict):
    del minimal_swagger_dict["definitions"]
    validate_spec(minimal_swagger_dict)


def test_empty_definitions_success(minimal_swagger_dict):
    validate_spec(minimal_swagger_dict)


def test_api_parameters_as_refs():
    # Verify issue #29 - instragram.json comes from:
    #
    # http://editor.swagger.io/#/
    #    -> File
    #       -> Open Example...
    #          -> instagram.yaml
    #
    # and then export it to a json file.
    instagram_specs, _ = get_spec_json_and_url("./tests/data/v2.0/instagram.json")
    validate_spec(instagram_specs)


def test_fails_on_invalid_external_ref_in_dict():
    # The external ref in petstore.json is valid.
    # The contents of the external ref (pet.json#/getall) is not - the 'name'
    # key in the parameter is missing.

    petstore_spec, petstore_url = get_spec_json_and_url(
        "./tests/data/v2.0/test_fails_on_invalid_external_ref/petstore.json"
    )

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(petstore_spec, petstore_url)

    assert "is not valid under any of the given schemas" in str(excinfo.value)


def test_fails_on_invalid_external_ref_in_list():
    # The external ref in petstore.json is valid.
    # The contents of the external ref (pet.json#/get_all_parameters) is not
    # - the 'name' key in the parameter is missing.
    petstore_spec, petstore_url = get_spec_json_and_url(
        "./tests/data/v2.0/test_fails_on_invalid_external_ref_in_list/petstore.json"
    )

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(petstore_spec, petstore_url)

    assert "is not valid under any of the given schemas" in str(excinfo.value)


@pytest.fixture
def node_spec():
    """Used in tests that have recursive $refs"""
    return {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "child": {
                "$ref": "#/definitions/Node",
            },
        },
        "required": ["name"],
    }


def test_recursive_ref(minimal_swagger_dict, node_spec):
    minimal_swagger_dict["definitions"]["Node"] = node_spec
    validate_spec(minimal_swagger_dict)


def test_recursive_ref_failure(minimal_swagger_dict, node_spec):
    minimal_swagger_dict["definitions"]["Node"] = node_spec
    # insert non-existent $ref
    node_spec["properties"]["foo"] = {"$ref": "#/definitions/Foo"}
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(minimal_swagger_dict)
    assert "Unresolvable JSON pointer" in str(excinfo.value)


def test_complicated_refs():
    # Split the swagger spec into a bunch of different json files and use
    # $refs all over to place to wire stuff together - see the test-data
    # files or this will make no sense whatsoever.
    file_path = "./tests/data/v2.0/test_complicated_refs/swagger.json"
    swagger_dict, origin_url = get_spec_json_and_url(file_path)
    resolver = validate_spec(swagger_dict, spec_url=origin_url)

    # Hokey verification but better than nothing:
    #   If all the files with $refs were ingested and validated and an
    #   exception was not thrown, there should be 7 cached file references
    #   in the resolver's store:
    #       6 json files from tests/data/v2.0/tests_complicated_refs/*
    #       1 yaml files from tests/data/v2.0/tests_complicated_refs/*
    assert len([uri for uri in resolver.store.keys() if uri.startswith("file://")]) == 7


def test_specs_with_discriminator():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    validate_spec(swagger_dict)


def test_specs_with_discriminator_fail_because_not_required():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["GenericPet"]["discriminator"] = "name"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (name) must be a required property" in str(excinfo.value)


def test_specs_with_discriminator_fail_because_not_string():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["GenericPet"]["discriminator"] = "weight"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (weight) must be a string property" in str(excinfo.value)


def test_specs_with_discriminator_fail_because_not_in_properties():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["GenericPet"]["discriminator"] = "an_other_property"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (an_other_property) must be defined in properties" in str(
        excinfo.value
    )


def test_specs_with_discriminator_in_allOf():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    validate_spec(swagger_dict)


def test_specs_with_discriminator_in_allOf_fail_because_not_required():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["BaseObject"]["discriminator"] = "name"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (name) must be a required property" in str(excinfo.value)


def test_specs_with_discriminator_in_allOf_fail_because_not_string():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["BaseObject"]["discriminator"] = "weight"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (weight) must be a string property" in str(excinfo.value)


def test_specs_with_discriminator_in_allOf_fail_because_not_in_properties():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["BaseObject"]["discriminator"] = "an_other_property"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (an_other_property) must be defined in properties" in str(
        excinfo.value
    )


def test_read_yaml_specs():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    swagger_dict["definitions"]["BaseObject"]["discriminator"] = "an_other_property"

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(swagger_dict)
    assert "discriminator (an_other_property) must be defined in properties" in str(
        excinfo.value
    )


@pytest.fixture
def default_checks_spec_dict(minimal_swagger_dict):
    minimal_swagger_dict["definitions"]["bool_string"] = {
        "properties": {
            "value": {
                "type": "string",
                "enum": ["True", "False"],
            },
        },
    }
    return minimal_swagger_dict


@pytest.mark.parametrize(
    "property_spec",
    [
        {"type": "integer", "default": 1},
        {"type": "boolean", "default": True},
        {"type": "null", "default": None},
        {"type": "number", "default": 2},
        {"type": "number", "default": 3.4},
        {"type": "object", "default": {"a_random_property": "valid"}},
        {"type": "array", "items": {"type": "integer"}, "default": [5, 6, 7]},
        {"type": "string", "default": ""},
        {"type": "string", "default": None, "x-nullable": True},
        {"default": -1},  # if type is not defined any value is a valid value
        {
            "type": "array",
            "items": {"$ref": "#/definitions/bool_string"},
            "default": [{"value": "False"}],
        },
    ],
)
def test_valid_specs_with_check_of_default_types(
    default_checks_spec_dict, property_spec
):
    default_checks_spec_dict["definitions"]["injected_definition"] = {
        "properties": {"property": property_spec},
    }
    # Success if no exception are raised
    validate_spec(default_checks_spec_dict)


@pytest.mark.parametrize(
    "property_spec, validator, instance",
    [
        [
            {"type": "integer", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "boolean", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "null", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "number", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "object", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "array", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "string", "default": -1},
            "type",
            -1,
        ],
        [
            {"type": "string", "minLength": 100, "default": "short_string"},
            "minLength",
            "short_string",
        ],
        [
            {"type": ["number", "boolean"], "default": "not_a_number_or_boolean"},
            "type",
            "not_a_number_or_boolean",
        ],
        [
            {
                "type": "array",
                "items": {"$ref": "#/definitions/bool_string"},
                "default": [{"value": "not_valid"}],
            },
            "enum",
            "not_valid",
        ],
    ],
)
def test_failure_due_to_wrong_default_type(
    default_checks_spec_dict, property_spec, validator, instance
):
    default_checks_spec_dict["definitions"]["injected_definition"] = {
        "properties": {"property": property_spec},
    }
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(default_checks_spec_dict)

    validation_error = excinfo.value.args[1]
    assert validation_error.instance == instance
    assert validation_error.validator == validator


def test_ref_without_str_argument(minimal_swagger_dict):
    not_a_ref = {"properties": {"$ref": {"type": "string"}}}
    minimal_swagger_dict["definitions"]["not_a_ref"] = not_a_ref
    validate_spec(minimal_swagger_dict)


def test_failure_because_references_in_operation_responses():
    with open(
        "./tests/data/v2.0/invalid_swagger_specs_because_ref_in_responses.json"
    ) as f:
        invalid_spec = json.load(f)
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(invalid_spec)
    assert (
        "GET /endpoint does not have a valid responses section. "
        "That section cannot be just a reference to another object."
        in str(excinfo.value)
    )


def test_type_array_with_items_succeed_validation(minimal_swagger_dict):
    minimal_swagger_dict["definitions"] = {
        "definition_1": {
            "type": "array",
            "items": {
                "type": "string",
            },
        },
    }

    # Success if no exception are raised
    validate_spec(minimal_swagger_dict)


@pytest.mark.parametrize(
    "swagger_dict_override",
    (
        {
            "definitions": {
                "definition_1": {
                    "type": "array",
                },
            },
        },
    ),
)
def test_type_array_without_items_succeed_fails(
    minimal_swagger_dict, swagger_dict_override
):
    minimal_swagger_dict.update(swagger_dict_override)
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec(minimal_swagger_dict)

    assert (
        str(excinfo.value)
        == "Definition of type array must define `items` property (definition #/definitions/definition_1)."
    )


INVALID_SCHEMA_OBJECT = {
    "type": "object",
    "properties": {"prop": {"type": "string", "default": 1}},
}


@pytest.mark.parametrize(
    "swagger_dict_override",
    (
        {
            "definitions": {"definition_1": INVALID_SCHEMA_OBJECT},
        },
        {
            "definitions": {
                "definition_2": {
                    "type": "object",
                    "properties": {"prop": INVALID_SCHEMA_OBJECT},
                }
            },
        },
        {
            "parameters": {
                "param_body": {
                    "in": "body",
                    "name": "body",
                    "schema": INVALID_SCHEMA_OBJECT,
                },
            },
        },
        {
            "paths": {
                "/endpoint": {
                    "get": {
                        "parameters": [
                            {
                                "in": "body",
                                "name": "body",
                                "schema": INVALID_SCHEMA_OBJECT,
                            }
                        ],
                        "responses": {
                            "200": {"description": "desc"},
                        },
                    },
                },
            },
        },
        {
            "paths": {
                "/endpoint": {
                    "get": {
                        "responses": {
                            "200": {
                                "description": "desc",
                                "schema": INVALID_SCHEMA_OBJECT,
                            },
                        },
                    },
                },
            },
        },
    ),
)
def test_highlight_inconsistent_schema_object_validation(
    minimal_swagger_dict, swagger_dict_override
):
    minimal_swagger_dict.update(swagger_dict_override)
    with pytest.raises(SwaggerValidationError):
        validate_spec(minimal_swagger_dict)


def test_additional_properties_validated(
    minimal_swagger_dict, default_checks_spec_dict
):
    invalid_spec = {"type": "object", "required": ["foo"]}
    minimal_swagger_dict["definitions"]["injected_definition"] = {
        "type": "object",
        "additionalProperties": invalid_spec,
    }

    with pytest.raises(SwaggerValidationError):
        validate_spec(minimal_swagger_dict)


def test_additional_properties_bool(minimal_swagger_dict, default_checks_spec_dict):
    minimal_swagger_dict["definitions"]["injected_definition"] = {
        "type": "object",
        "additionalProperties": False,
    }

    validate_spec(minimal_swagger_dict)


def test_array_items_validated(minimal_swagger_dict, default_checks_spec_dict):
    invalid_spec = {"type": "object", "required": ["foo"]}
    minimal_swagger_dict["definitions"]["injected_definition"] = {
        "type": "array",
        "items": invalid_spec,
    }

    with pytest.raises(SwaggerValidationError):
        validate_spec(minimal_swagger_dict)



import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator20 import validate_definitions


@pytest.mark.parametrize(
    "property_spec",
    [
        {"type": "integer", "default": 1},
        {"type": "boolean", "default": True},
        {"type": "null", "default": None},
        {"type": "number", "default": 2},
        {"type": "number", "default": 3.4},
        {"type": "object", "default": {"a_random_property": "valid"}},
        {"type": "array", "items": {"type": "integer"}, "default": [5, 6, 7]},
        {"default": -1},  # if type is not defined any value is a valid value
        {"type": "string", "default": ""},
    ],
)
def test_api_check_default_succeed(property_spec):
    definitions = {
        "injected_definition": {
            "properties": {
                "property": property_spec,
            },
        },
    }

    # Success if no exception are raised
    validate_definitions(definitions, lambda x: x)


@pytest.mark.parametrize(
    "property_spec, validator, instance",
    [
        [
            {"type": "integer", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "boolean", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "null", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "number", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "object", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "array", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "string", "default": -1},
            "type",
            -1,
        ],
        [
            {"type": "string", "minLength": 100, "default": "short_string"},
            "minLength",
            "short_string",
        ],
    ],
)
def test_api_check_default_fails(property_spec, validator, instance):
    definitions = {
        "injected_definition": {
            "properties": {
                "property": property_spec,
            },
        },
    }

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_definitions(definitions, lambda x: x)

    validation_error = excinfo.value.args[1]
    assert validation_error.instance == instance
    assert validation_error.validator == validator


def test_multiple_types_fail():
    definitions = {
        "definition_1": {
            "type": ["number", "boolean"],
        },
    }

    with pytest.raises(
        SwaggerValidationError,
        match=r"In definition of .*, type must be a string; lists are not allowed \(.*\)",
    ) as excinfo:
        validate_definitions(definitions, lambda x: x)

    assert str(definitions["definition_1"]["type"]) in str(excinfo.value)


def test_type_array_with_items_succeed_validation():
    definitions = {
        "definition_1": {
            "type": "array",
            "items": {
                "type": "string",
            },
        },
    }

    # Success if no exception are raised
    validate_definitions(definitions, lambda x: x)


def test_type_array_without_items_succeed_fails():
    definitions = {
        "definition_1": {
            "type": "array",
        },
    }

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_definitions(definitions, lambda x: x)

    assert (
        str(excinfo.value)
        == "Definition of type array must define `items` property (definition #/definitions/definition_1)."
    )


def test_inline_model_is_not_valid_validation_fails():
    definitions = {
        "definition_1": {
            "properties": {
                "property": {
                    "type": "array",
                },
            },
        },
    }

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_definitions(definitions, lambda x: x)

    assert (
        str(excinfo.value) == "Definition of type array must define `items` property "
        "(definition #/definitions/definition_1/properties/property)."
    )



from __future__ import annotations

import functools
import json
import logging
import string
import warnings
from typing import Any
from typing import Callable
from typing import cast
from typing import Iterable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem
    from jsonschema.validators import _Handler
from jsonschema.validators import Draft4Validator
from jsonschema.validators import RefResolver

from swagger_spec_validator import ref_validators
from swagger_spec_validator.common import get_uri_from_file_path
from swagger_spec_validator.common import read_resource_file
from swagger_spec_validator.common import read_url
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.common import SwaggerValidationWarning
from swagger_spec_validator.common import wrap_exception
from swagger_spec_validator.ref_validators import default_handlers
from swagger_spec_validator.ref_validators import in_scope
from swagger_spec_validator.ref_validators import validate_schema_value


log = logging.getLogger(__name__)


def validate_ref(ref_dict: dict[str, Any], path: list[str]) -> None:
    """Check if a ref_dict has siblings that will be overwritten by $ref or $ref is None.

    While the siblings case does not contradict the spec, it may cause confusion and mislead
    developers. See https://stackoverflow.com/a/48114924.

    :param ref_dict: A dict that may be {'$ref': '#/blah/blah', 'x-nullable': true}.
    :type ref_dict: dict

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    keys_to_ignore = {"x-scope", "$ref", "description"}

    if any(key for key in ref_dict.keys() if key not in keys_to_ignore):
        warnings.warn(
            SwaggerValidationWarning(
                'Found "$ref: {}" with siblings that will be overwritten. '
                "See https://stackoverflow.com/a/48114924 for more information. (path {})".format(
                    ref_dict["$ref"],
                    "/".join(path),
                )
            ),
        )

    if ref_dict["$ref"] is None:
        warnings.warn(
            SwaggerValidationWarning(
                "Identified $ref with None value. This is usually an error, "
                "although technically it might be allowed. (path: {})".format(
                    "/".join(path)
                ),
            ),
        )


def validate_references(
    raw_spec: dict[Any, Any] | list[Any],
    deref: Callable,
    path: list[str] | None = None,
    visited_spec_ids: set[int] | None = None,
) -> None:
    if path is None:
        path = ["#"]

    if visited_spec_ids is None:
        visited_spec_ids = set()

    if id(raw_spec) in visited_spec_ids:
        return

    # Register raw_spec into visited_spec_ids to prevent unbounded recursion
    visited_spec_ids.add(id(raw_spec))

    if isinstance(raw_spec, dict) and "$ref" in raw_spec:
        # The additional check is needed as `is_ref` will consider raw_spec dictionaries with `$ref`
        # attribute and string value.
        # The goal of validate_ref is to ensure that objects that looks like a reference is actually
        # valid by warning users about those cases.
        # Due to _looks like_ we need to almost duplicate the check by removing the string enforcement
        validate_ref(ref_dict=raw_spec, path=path)

    if is_ref(raw_spec) and "x-scope" in raw_spec:
        # Ensure that we're following references only if they were already
        # descended by jsonschema during initial schema validation (x-scope is present)
        # This is mostly done to ensure that we're not causing RefResolutionError
        # due to the fact that the reference is relative and we have no information
        # about the base path.
        # This could be handled differently but we thought that checking references that
        # were actively part of the specs is a valid trade-off here
        validate_references(deref(raw_spec), deref, path, visited_spec_ids)
    elif isinstance(raw_spec, dict):
        for k, v in sorted(raw_spec.items()):
            validate_references(v, deref, path + [k], visited_spec_ids)
    elif isinstance(raw_spec, list):
        for k, v in enumerate(raw_spec):
            validate_references(v, deref, path + [str(k)], visited_spec_ids)


def deref(
    ref_dict: dict[Any, Any], resolver: RefResolver
) -> int | float | None | bool | list[Any] | dict[Any, Any]:
    """Dereference ref_dict (if it is indeed a ref) and return what the
    ref points to.

    :param ref_dict: Something like {'$ref': '#/blah/blah'}
    :type ref_dict: dict
    :param resolver: Ref resolver used to do the de-referencing
    :type resolver: :class:`jsonschema.RefResolver`

    :return: de-referenced value of ref_dict
    :rtype: scalar, list, dict
    """
    if ref_dict is None or not is_ref(ref_dict):
        return ref_dict

    ref = ref_dict["$ref"]
    with in_scope(resolver, ref_dict):
        with resolver.resolving(ref) as target:
            log.debug("Resolving %s", ref)
            return target


@wrap_exception
def validate_spec_url(spec_url: str) -> RefResolver:
    """Validates a Swagger 2.0 API Specification at the given URL.

    :param spec_url: the URL of the service's swagger spec.

    :returns: The resolver (with cached remote refs) used during validation
    :rtype: :class:`jsonschema.RefResolver`
    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    log.info("Validating %s", spec_url)
    return validate_spec(read_url(spec_url), spec_url)


def validate_spec(
    spec_dict: dict[Any, Any],
    spec_url: str = "",
    http_handlers: SupportsKeysAndGetItem[str, _Handler]
    | Iterable[tuple[str, _Handler]]
    | None = None,
) -> RefResolver:
    """Validates a Swagger 2.0 API Specification given a Swagger Spec.

    :param spec_dict: the json dict of the swagger spec.
    :type spec_dict: dict
    :param spec_url: url from which spec_dict was retrieved. Used for
        dereferencing refs. eg: file:///foo/swagger.json
    :type spec_url: string
    :param http_handlers: used to download any remote $refs in spec_dict with
        a custom http client. Defaults to None in which case the default
        http client built into jsonschema's RefResolver is used. This
        is a mapping from uri scheme to a callable that takes a
        uri.

    :returns: the resolver (with cached remote refs) used during validation
    :rtype: :class:`jsonschema.RefResolver`
    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    swagger_resolver = validate_json(
        spec_dict,
        "schemas/v2.0/schema.json",
        spec_url=spec_url,
        http_handlers=http_handlers,
    )

    bound_deref = functools.partial(deref, resolver=swagger_resolver)
    spec_dict = cast("dict[Any, Any]", bound_deref(spec_dict))
    apis = bound_deref(spec_dict["paths"])
    definitions = bound_deref(spec_dict.get("definitions", {}))
    validate_apis(cast("dict[Any, Any]", apis), bound_deref)
    validate_definitions(cast("dict[Any, Any]", definitions), bound_deref)
    validate_parameters(
        cast("dict[Any, Any]", bound_deref(spec_dict.get("parameters", {}))),
        bound_deref,
    )
    validate_references(spec_dict, bound_deref)
    return swagger_resolver


@wrap_exception
def validate_json(
    spec_dict: list[Any] | dict[str, Any],
    schema_path: str,
    spec_url: str = "",
    http_handlers: SupportsKeysAndGetItem[str, _Handler]
    | Iterable[tuple[str, _Handler]]
    | None = None,
) -> RefResolver:
    """Validate a json document against a json schema.

    :param spec_dict: json document in the form of a list or dict.
    :param schema_path: package relative path of the json schema file.
    :param spec_url: base uri to use when creating a
        RefResolver for the passed in spec_dict.
    :param http_handlers: used to download any remote $refs in spec_dict with
        a custom http client. Defaults to None in which case the default
        http client built into jsonschema's RefResolver is used. This
        is a mapping from uri scheme to a callable that takes a
        uri.

    :return: RefResolver for spec_dict with cached remote $refs used during
        validation.
    :rtype: :class:`jsonschema.RefResolver`
    """
    schema, schema_path = read_resource_file(schema_path)

    schema_resolver = RefResolver(
        base_uri=get_uri_from_file_path(schema_path),
        referrer=schema,
        handlers=default_handlers,
    )

    spec_resolver = RefResolver(
        base_uri=spec_url,
        referrer=cast("dict[str, Any]", spec_dict),
        handlers=http_handlers or default_handlers,
    )

    ref_validators.validate(
        instance=spec_dict,
        schema=schema,
        resolver=schema_resolver,
        instance_cls=ref_validators.create_dereffing_validator(spec_resolver),
        cls=Draft4Validator,
    )

    # Since remote $refs were downloaded, pass the resolver back to the caller
    # so that its cached $refs can be re-used.
    return spec_resolver


@wrap_exception
def validate_value_type(schema: dict[str, Any], value: Any, deref: Callable) -> None:
    # Extract resolver from deref partial build on ``validate_spec``
    # This is used in order to use already fetched external references
    # If it is missing a new RefResolver will be initialized
    swagger_resolver = getattr(deref, "keywords", {}).get("resolver", None)
    validate_schema_value(
        schema=deref(schema), value=value, swagger_resolver=swagger_resolver
    )


def validate_default_in_parameter(param_spec: dict[str, Any], deref: Callable) -> None:
    deref_param_spec = deref(param_spec)

    if deref_param_spec.get("required"):
        # If the parameter is a required parameter, default has no meaning
        return

    if "default" in deref_param_spec:
        if (
            deref_param_spec["default"] is None
            and deref_param_spec.get("x-nullable", False) is True
        ):
            # In case x-nullable property is set to true, null is a valid default
            return
        validate_value_type(
            schema=deref_param_spec,
            value=deref_param_spec["default"],
            deref=deref,
        )


def validate_defaults_in_parameters(params_spec: list[Any], deref: Callable) -> None:
    """
    Validates that default values for api parameters are
    of the parameter type

    :param params_spec: list of parameter objects (#/paths/<path>/<http_verb>/parameters)
    :param deref: callable that dereferences $refs

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    # Note: this functions is preserved to avoid public signature updates (it's not used internally)
    for param_spec in params_spec:
        validate_default_in_parameter(param_spec, deref)


def validate_responses(
    api: str, http_verb: str, responses_dict: dict[str, Any], deref: Callable
) -> None:
    if is_ref(responses_dict):
        raise SwaggerValidationError(
            "{http_verb} {api} does not have a valid responses section. "
            "That section cannot be just a reference to another object.".format(
                http_verb=http_verb.upper(),
                api=api,
            )
        )
    for response_status, response_spec in responses_dict.items():
        response_schema = response_spec.get("schema")
        if response_schema is None:
            continue
        validate_definition(
            definition=response_schema,
            deref=deref,
            def_name="#/paths/{api}/{http_verb}/responses/{status_code}".format(
                http_verb=http_verb,
                api=api,
                status_code=response_status,
            ),
            visited_definitions=set(),
        )


def validate_non_body_parameter(
    param: dict[str, Any], deref: Callable, def_name: str
) -> None:
    if "type" not in param:
        raise SwaggerValidationError(
            "Non-Body parameter in `{def_name}` does not specify `type`.".format(
                def_name=def_name
            ),
        )

    if param["type"] == "array" and "items" not in param:
        raise SwaggerValidationError(
            "Non-Body array parameter in `{def_name}` does not specify `items`.".format(
                def_name=def_name
            ),
        )


def validate_body_parameter(
    param: dict[str, Any], deref: Callable, def_name: str
) -> None:
    if "schema" not in param:
        raise SwaggerValidationError(
            "Body parameter in `{def_name}` does not specify `schema`.".format(
                def_name=def_name
            )
        )

    validate_definition(
        definition=param["schema"],
        deref=deref,
        def_name=f"{def_name}/schema",
        visited_definitions=set(),
    )


def validate_parameter(param: dict[str, Any], deref: Callable, def_name: str) -> None:
    validate_default_in_parameter(param, deref)

    if not is_ref(param):
        if param["in"] == "body":
            validate_body_parameter(param, deref, def_name)
        else:
            validate_non_body_parameter(param, deref, def_name)


def validate_apis(apis: dict[str, Any], deref: Callable) -> None:
    """Validates semantic errors in #/paths.

    :param apis: dict of all the #/paths
    :param deref: callable that dereferences $refs

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    :raises: :py:class:`jsonschema.exceptions.ValidationError`
    """
    operation_id_set = set()

    for api_name, api_body in apis.items():
        api_body = deref(api_body)
        api_params = deref(api_body.get("parameters", []))
        validate_duplicate_param(api_params, deref)
        for idx, param in enumerate(api_params):
            validate_parameter(
                param=param,
                deref=deref,
                def_name="#/paths/{api_name}/parameters/{idx}".format(
                    api_name=api_name,
                    idx=idx,
                ),
            )

        for oper_name in api_body:
            # don't treat parameters that apply to all api operations as
            # an operation
            if oper_name == "parameters" or oper_name.startswith("x-"):
                continue
            oper_body = deref(api_body[oper_name])

            # Check that, if this operation has an operationId defined,
            # no other operation also has that operationId.
            operation_id = oper_body.get("operationId")
            if operation_id is not None:
                if operation_id in operation_id_set:
                    raise SwaggerValidationError(
                        f"Duplicate operationId: {operation_id}"
                    )
                operation_id_set.add(operation_id)

            oper_params = deref(oper_body.get("parameters", []))
            validate_duplicate_param(oper_params, deref)
            all_path_params = list(
                set(
                    get_path_param_names(api_params, deref)
                    + get_path_param_names(oper_params, deref),
                )
            )
            validate_unresolvable_path_params(api_name, all_path_params)
            for idx, param in enumerate(oper_params):
                validate_parameter(
                    param=param,
                    deref=deref,
                    def_name="#/paths/{api_name}/{oper_name}/parameters/{idx}".format(
                        api_name=api_name,
                        oper_name=oper_name,
                        idx=idx,
                    ),
                )
            # Responses validation
            validate_responses(api_name, oper_name, oper_body["responses"], deref)


def get_collapsed_properties_type_mappings(
    definition: dict[str, Any], deref: Callable
) -> tuple[dict[Any, Any], dict[Any, Any]]:
    """
    Get all the properties for a swagger model (definition).
    :param definition: dictionary representation of the definition
    :type definition: dict
    :param deref: callable that dereferences $refs
    :return: (required properties type mapping, not required properties type mapping)
    :type: tuple
    """
    definition = deref(definition)
    required_properties = {}
    not_required_properties = {}
    if definition.get("allOf"):
        for inner_definition in definition["allOf"]:
            (
                inner_required_properties,
                inner_not_required_properties,
            ) = get_collapsed_properties_type_mappings(inner_definition, deref)
            required_properties.update(inner_required_properties)
            not_required_properties.update(inner_not_required_properties)
    else:
        properties = {
            prop_name: prop_schema.get("type", "object")
            for prop_name, prop_schema in definition.get("properties", {}).items()
        }
        required_properties_set = set(definition.get("required", []))
        for k, v in properties.items():
            if k in required_properties_set:
                required_properties[k] = v
            else:
                not_required_properties[k] = v

    return required_properties, not_required_properties


def validate_property_default(property_spec: dict[str, Any], deref: Callable) -> None:
    """
    Validates that default values for definitions are of the property type.
    Enforces presence of "type" in case of deffefwf"default" presence.

    :param property_spec: schema object (#/definitions/<def_name>/properties/<property_name>
    :param deref: callable that dereferences $refs

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    deref_property_spec = deref(property_spec)
    if "default" in deref_property_spec:
        if (
            deref_property_spec["default"] is None
            and deref_property_spec.get("x-nullable", False) is True
        ):
            # In case x-nullable property is set to true, null is a valid default
            return

        validate_value_type(
            schema=property_spec, value=deref_property_spec["default"], deref=deref
        )


def validate_defaults_in_definition(
    definition_spec: dict[str, Any], deref: Callable
) -> None:
    for property_name, property_spec in (definition_spec.get("properties", {})).items():
        validate_property_default(property_spec, deref)


def validate_arrays_in_definition(
    definition_spec: dict[str, Any],
    deref: Callable,
    def_name: str | None = None,
    visited_definitions: set[str] | None = None,
) -> None:
    if definition_spec.get("type") == "array":
        if "items" not in definition_spec:
            raise SwaggerValidationError(
                "Definition of type array must define `items` property{}.".format(
                    "" if not def_name else f" (definition {def_name})",
                ),
            )
        validate_definition(
            definition=definition_spec["items"],
            deref=deref,
            def_name=f"{def_name}/items",
            visited_definitions=visited_definitions,
        )


def validate_definition(
    definition: dict[str, Any],
    deref: Callable,
    def_name: str | None = None,
    visited_definitions: set[str] | None = None,
) -> None:
    """
    :param visited_definitions: set of already visited definitions
                                    This is used to cut recursion in case of recursive definitions
    :type visited_definitions: set
    """
    if visited_definitions is not None:
        # Remove x-scope or else no two definitions will be the same
        stripped_definition = json.dumps(
            {key: definition[key] for key in definition if key != "x-scope"},
            sort_keys=True,
        )
        if stripped_definition in visited_definitions:
            return
        visited_definitions.add(stripped_definition)

    definition = deref(definition)

    swagger_type = definition.get("type")
    if isinstance(swagger_type, list):
        # not valid Swagger; see https://github.com/OAI/OpenAPI-Specification/issues/458
        raise SwaggerValidationError(
            "In definition of {}, type must be a string; lists are not allowed ({})".format(
                def_name or "(no name)", swagger_type
            )
        )

    if "allOf" in definition:
        for idx, inner_definition in enumerate(definition["allOf"]):
            validate_definition(
                definition=inner_definition,
                deref=deref,
                def_name=f"{def_name}/{str(idx)}",
                visited_definitions=visited_definitions,
            )
    else:
        required = definition.get("required", [])
        props = definition.get("properties", {}).keys()
        extra_props = list(set(required) - set(props))
        if extra_props:
            raise SwaggerValidationError(
                "In definition of {}, required list has properties not defined: {}.".format(
                    def_name or "(no name)",
                    extra_props,
                )
            )

        validate_defaults_in_definition(definition, deref)
        validate_arrays_in_definition(
            definition_spec=definition,
            deref=deref,
            def_name=def_name,
            visited_definitions=visited_definitions,
        )

        for property_name, property_spec in definition.get("properties", {}).items():
            validate_definition(
                definition=property_spec,
                deref=deref,
                def_name=f"{def_name}/properties/{property_name}",
                visited_definitions=visited_definitions,
            )

    if "additionalProperties" in definition:
        if definition.get("additionalProperties") not in (True, False):
            validate_definition(
                definition=cast("dict[str, Any]", definition["additionalProperties"]),
                deref=deref,
                def_name=f"{def_name}/additionalProperties",
                visited_definitions=visited_definitions,
            )

    if "discriminator" in definition:
        required_props, not_required_props = get_collapsed_properties_type_mappings(
            definition, deref
        )
        discriminator = definition["discriminator"]
        if (
            discriminator not in required_props
            and discriminator not in not_required_props
        ):
            raise SwaggerValidationError(
                "In definition of {}, discriminator ({}) must be defined in properties".format(
                    def_name or "(no name)", discriminator
                )
            )
        if discriminator not in required_props:
            raise SwaggerValidationError(
                "In definition of {}, discriminator ({}) must be a required property".format(
                    def_name or "(no name)", discriminator
                )
            )
        if required_props[discriminator] != "string":
            raise SwaggerValidationError(
                "In definition of {}, discriminator ({}) must be a string property".format(
                    def_name or "(no name)", discriminator
                )
            )


def validate_definitions(definitions: dict[str, Any], deref: Callable) -> None:
    """Validates the semantic errors in #/definitions.

    :param definitions: dict of all the definitions
    :param deref: callable that dereferences $refs

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    :raises: :py:class:`jsonschema.exceptions.ValidationError`
    """
    visited_definitions: set[str] = set()
    for def_name, definition in definitions.items():
        validate_definition(
            definition=definition,
            deref=deref,
            def_name=f"#/definitions/{def_name}",
            visited_definitions=visited_definitions,
        )


def validate_parameters(parameters: dict[str, Any], deref: Callable) -> None:
    for param_name, param_spec in parameters.items():
        validate_parameter(
            param=param_spec,
            deref=deref,
            def_name=f"#/parameters/{param_name}",
        )


def get_path_param_names(params: list[Any], deref: Callable) -> list[str]:
    """Fetch all the names of the path parameters of an operation.

    :param params: list of all the params
    :param deref: callable that dereferences $refs

    :returns: list of the name of the path params
    """
    return [deref(param)["name"] for param in params if deref(param)["in"] == "path"]


def validate_duplicate_param(params: list[Any], deref: Callable) -> None:
    """Validate no duplicate parameters are present.

    Uniqueness is determined by the tuple ('name', 'in').

    :param params: list of all the params
    :param deref: callable that dereferences $refs

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError` when
        a duplicate parameter is found.
    """
    seen = set()
    msg = "Duplicate param found with (name, in)"
    for param in params:
        param = deref(param)
        param_key = (param["name"], param["in"])
        if param_key in seen:
            raise SwaggerValidationError(f"{msg}: {param_key}")
        seen.add(param_key)


def get_path_params_from_url(path: str) -> list[str]:
    """Parse the path parameters from a path string

    :param path: path url to parse for parameters

    :returns: List of path parameter names
    """
    formatter = string.Formatter()
    return [item[1] for item in formatter.parse(path) if item[1]]


def validate_unresolvable_path_params(path_name: str, path_params: list[str]) -> None:
    """Validate that every path parameter listed is also defined.

    :param path_name: complete path name as a string.
    :param path_params: Names of all the eligible path parameters

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    for path in get_path_params_from_url(path_name):
        if path not in path_params:
            msg = "Path parameter '{}' used is not documented on '{}'".format(
                path, path_name
            )
            raise SwaggerValidationError(msg)


def is_ref(spec_dict: dict[Any, Any] | list[Any]) -> bool:
    return isinstance(spec_dict, dict) and isinstance(spec_dict.get("$ref"), str)



import json
import os

import pytest

from swagger_spec_validator.common import get_uri_from_file_path
from tests import TESTS_BASE_PATH


@pytest.fixture(scope="session")
def petstore_contents():
    with open(TESTS_BASE_PATH + "/data/v2.0/petstore.json") as f:
        return f.read()


@pytest.fixture
def petstore_dict(petstore_contents):
    return json.loads(petstore_contents)


def get_spec_json_and_url(rel_path):
    abs_path = os.path.abspath(rel_path)
    with open(abs_path) as f:
        return json.loads(f.read()), get_uri_from_file_path(abs_path)



from unittest import mock

import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.util import validate_spec_url
from tests.conftest import is_urlopen_error


def test_raise_SwaggerValidationError_on_urlopen_error():
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec_url("http://foo")
    assert is_urlopen_error(excinfo.value)


@mock.patch("swagger_spec_validator.util.read_url")
@mock.patch("swagger_spec_validator.util.get_validator")
def test_validate_spec_url_success(mock_get_validator, mock_read_url):
    spec_url = mock.Mock()
    validate_spec_url(spec_url)
    mock_read_url.assert_called_once_with(spec_url)
    mock_get_validator.assert_called_once_with(mock_read_url.return_value, spec_url)



__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
]

__title__ = "swagger-spec-validator"
__summary__ = "Validation of Swagger specifications"
__uri__ = "http://github.com/Yelp/swagger_spec_validator"

__version__ = "3.0.4"

__author__ = "Yelp"
__email__ = "core-services@yelp.com"

__license__ = "Apache License, Version 2.0"



import json
import os
from unittest import mock

import pytest

from swagger_spec_validator.common import get_uri_from_file_path
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.common import SwaggerValidationWarning
from swagger_spec_validator.validator20 import validate_spec_url
from tests.conftest import is_urlopen_error


def test_success(petstore_contents):
    with mock.patch(
        "swagger_spec_validator.validator20.read_url",
        return_value=json.loads(petstore_contents),
    ) as mock_read_url:
        validate_spec_url("http://localhost/api-docs")
        mock_read_url.assert_called_once_with("http://localhost/api-docs")


def test_success_crossref_url_yaml():
    urlpath = get_uri_from_file_path(os.path.abspath("./tests/data/v2.0/minimal.yaml"))
    validate_spec_url(urlpath)


def test_success_crossref_url_json():
    urlpath = get_uri_from_file_path(
        os.path.abspath("./tests/data/v2.0/relative_ref.json")
    )
    validate_spec_url(urlpath)


def test_complicated_refs_json():
    urlpath = get_uri_from_file_path(
        os.path.abspath("./tests/data/v2.0/test_complicated_refs/swagger.json")
    )
    validate_spec_url(urlpath)


def test_specs_with_empty_reference():
    with pytest.warns(SwaggerValidationWarning) as warninfo:
        validate_spec_url(
            get_uri_from_file_path(
                os.path.abspath(
                    "./tests/data/v2.0/invalid_swagger_spec_because_empty_reference.yaml"
                ),
            ),
        )

    assert (
        "Identified $ref with None value. This is usually an error, although technically it might be allowed. "
        "(path: #/definitions/model1/x-extends)" == str(warninfo.list[0].message)
    )


def test_raise_SwaggerValidationError_on_urlopen_error():
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_spec_url("http://foo")
    assert is_urlopen_error(excinfo.value)



import functools

from swagger_spec_validator.validator20 import deref
from swagger_spec_validator.validator20 import get_collapsed_properties_type_mappings
from swagger_spec_validator.validator20 import validate_json
from tests.validator20.conftest import get_spec_json_and_url


def get_deref(spec_dict):
    swagger_resolver = validate_json(
        spec_dict,
        "schemas/v2.0/schema.json",
    )
    return functools.partial(deref, resolver=swagger_resolver)


def test_get_collapsed_properties_type_mapping_simple_case():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    (
        required_parameters,
        not_required_parameters,
    ) = get_collapsed_properties_type_mappings(
        definition=swagger_dict["definitions"]["GenericPet"],
        deref=get_deref(swagger_dict),
    )
    assert required_parameters == {"type": "string", "weight": "integer"}
    assert not_required_parameters == {"name": "string"}


def test_get_collapsed_properties_type_mapping_allOf_add_required_property():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    (
        required_parameters,
        not_required_parameters,
    ) = get_collapsed_properties_type_mappings(
        definition=swagger_dict["definitions"]["Dog"],
        deref=get_deref(swagger_dict),
    )
    assert required_parameters == {
        "type": "string",
        "weight": "integer",
        "birth_date": "string",
    }
    assert not_required_parameters == {"name": "string"}


def test_get_collapsed_properties_type_mapping_allOf_add_not_required_property():
    file_path = "./tests/data/v2.0/test_polymorphic_specs/swagger.json"
    swagger_dict, _ = get_spec_json_and_url(file_path)

    (
        required_parameters,
        not_required_parameters,
    ) = get_collapsed_properties_type_mappings(
        definition=swagger_dict["definitions"]["Cat"],
        deref=get_deref(swagger_dict),
    )
    assert required_parameters == {"type": "string", "weight": "integer"}
    assert not_required_parameters == {"name": "string", "color": "string"}



import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator20 import validate_apis
from swagger_spec_validator.validator20 import validate_defaults_in_parameters

RESPONSES = {
    "default": {
        "description": "random description",
    },
}


def test_api_level_params_ok():
    # Parameters defined at the API level apply to all operations within that
    # API. Make sure we don't treat the API level parameters as an operation
    # since they are peers.
    apis = {
        "/tags/{tag-name}": {
            "parameters": [
                {"name": "tag-name", "in": "path", "type": "string", "required": True},
            ],
            "get": {
                "responses": RESPONSES,
            },
        },
    }
    # Success == no exception thrown
    validate_apis(apis, lambda x: x)


def test_api_level_x_hyphen_ok():
    # Elements starting with "x-" should be ignored
    apis = {
        "/tags/{tag-name}": {
            "x-ignore-me": "DO NOT LOOK AT ME!",
            "get": {
                "parameters": [
                    {
                        "name": "tag-name",
                        "in": "path",
                        "type": "string",
                    }
                ],
                "responses": RESPONSES,
            },
        }
    }
    # Success == no exception thrown
    validate_apis(apis, lambda x: x)


@pytest.mark.parametrize(
    "partial_parameter_spec",
    [
        {"type": "integer", "default": 1},
        {"type": "boolean", "default": True},
        {"type": "null", "default": None},
        {"type": "number", "default": 2},
        {"type": "number", "default": 3.4},
        {"type": "object", "default": {"a_random_property": "valid"}},
        {"type": "array", "items": {"type": "integer"}, "default": [5, 6, 7]},
        {"type": "string", "default": ""},
        {"type": "string", "default": None, "x-nullable": True},
        {"type": ["number", "boolean"], "default": 8},
        {"type": ["number", "boolean"], "default": False},
    ],
)
def test_api_check_default_succeed(partial_parameter_spec):
    apis = {
        "/api": {
            "get": {
                "parameters": [
                    dict({"name": "param", "in": "query"}, **partial_parameter_spec),
                ],
                "responses": RESPONSES,
            },
        },
    }

    # Success if no exception are raised
    validate_apis(apis, lambda x: x)


@pytest.mark.parametrize(
    "partial_parameter_spec, validator, instance",
    [
        [
            {"type": "integer", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "boolean", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "null", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "number", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "object", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "array", "default": "wrong_type"},
            "type",
            "wrong_type",
        ],
        [
            {"type": "string", "default": -1},
            "type",
            -1,
        ],
        [
            {"type": "string", "minLength": 100, "default": "short_string"},
            "minLength",
            "short_string",
        ],
        [
            {"type": ["number", "boolean"], "default": "not_a_number_or_boolean"},
            "type",
            "not_a_number_or_boolean",
        ],
    ],
)
def test_api_check_default_fails(partial_parameter_spec, validator, instance):
    apis = {
        "/api": {
            "get": {
                "parameters": [
                    dict({"name": "param", "in": "query"}, **partial_parameter_spec),
                ],
                "responses": RESPONSES,
            },
        },
    }

    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_apis(apis, lambda x: x)

    validation_error = excinfo.value.args[1]
    assert validation_error.instance == instance
    assert validation_error.validator == validator


def test_validate_defaults_in_parameters_succeed():
    # Success if no exception are raised
    validate_defaults_in_parameters(
        params_spec=[{"type": "integer"}],
        deref=lambda x: x,
    )


def test_validate_defaults_in_parameters_fails():
    with pytest.raises(SwaggerValidationError):
        validate_defaults_in_parameters(
            params_spec=[
                {"type": "integer", "default": "wrong_type"},
            ],
            deref=lambda x: x,
        )


@pytest.mark.parametrize(
    "apis",
    [
        {
            "/api": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "responses": {},
                },
                "post": {
                    "operationId": "duplicateOperationId",
                    "responses": {},
                },
            },
        },
        {
            "/api1": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "responses": {},
                },
            },
            "/api2": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "responses": {},
                },
            },
        },
        {
            "/api1": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "tags": ["tag1", "tag2"],
                    "responses": {},
                },
            },
            "/api2": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "tags": ["tag1"],
                    "responses": {},
                },
            },
        },
    ],
)
def test_duplicate_operationIds_fails(apis):
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_apis(apis, lambda x: x)

    swagger_validation_error = excinfo.value
    error_message = swagger_validation_error.args[0]

    assert error_message == "Duplicate operationId: duplicateOperationId"


@pytest.mark.parametrize(
    "apis",
    [
        {
            "/api1": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "tags": ["tag1"],
                    "responses": {},
                },
            },
            "/api2": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "tags": ["tag2"],
                    "responses": {},
                },
            },
            "/api3": {
                "get": {
                    "operationId": "duplicateOperationId",
                    "responses": {},
                },
            },
        },
    ],
)
def test_duplicate_operationIds_fails_if_tags_differ(apis):
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_apis(apis, lambda x: x)

    swagger_validation_error = excinfo.value
    error_message = swagger_validation_error.args[0]

    assert error_message == "Duplicate operationId: duplicateOperationId"


def test_invalid_inline_models_in_responses_fails():
    apis = {
        "/endpoint": {
            "get": {
                "responses": {
                    "200": {
                        "description": "desc",
                        "schema": {
                            "type": "object",
                            "properties": {"prop": {"type": "array"}},
                        },
                    },
                },
            },
        },
    }
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_apis(apis, lambda x: x)
    assert (
        str(excinfo.value) == "Definition of type array must define `items` property "
        "(definition #/paths//endpoint/get/responses/200/properties/prop)."
    )


def test_invalid_inline_models_in_operation_body_parameters_fails():
    apis = {
        "/endpoint": {
            "get": {
                "parameters": [
                    {
                        "in": "body",
                        "name": "body",
                        "schema": {
                            "type": "object",
                            "properties": {"prop": {"type": "array"}},
                        },
                    }
                ],
                "responses": {
                    "200": {
                        "description": "desc",
                    },
                },
            },
        },
    }
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_apis(apis, lambda x: x)
    assert (
        str(excinfo.value) == "Definition of type array must define `items` property "
        "(definition #/paths//endpoint/get/parameters/0/schema/properties/prop)."
    )


def test_invalid_inline_models_in_api_body_parameters_fails():
    apis = {
        "/endpoint": {
            "parameters": [
                {
                    "in": "body",
                    "name": "body",
                    "schema": {
                        "type": "object",
                        "properties": {"prop": {"type": "array"}},
                    },
                }
            ],
            "get": {
                "responses": {
                    "200": {
                        "description": "desc",
                    },
                },
            },
        },
    }
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_apis(apis, lambda x: x)
    assert (
        str(excinfo.value) == "Definition of type array must define `items` property "
        "(definition #/paths//endpoint/parameters/0/schema/properties/prop)."
    )



from __future__ import annotations

import contextlib
import functools
import os
import sys
from functools import lru_cache
from typing import Any
from typing import Callable
from typing import TypeVar
from urllib.parse import urljoin
from urllib.request import pathname2url
from urllib.request import urlopen

import importlib_resources
import yaml
from typing_extensions import ParamSpec

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader  # type: ignore


TIMEOUT_SEC = 1.0
P = ParamSpec("P")
T = TypeVar("T")


def wrap_exception(method: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(method)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return method(*args, **kwargs)
        except Exception as e:
            raise SwaggerValidationError(str(e), e).with_traceback(sys.exc_info()[2])

    return wrapper


def get_uri_from_file_path(file_path: str) -> str:
    return urljoin("file://", pathname2url(os.path.abspath(file_path)))


def read_file(file_path: str) -> dict[str, Any]:
    """
    Utility method for reading a JSON/YAML file and converting it to a Python dictionary
    :param file_path: path of the file to read

    :return: Python dictionary representation of the JSON file
    :rtype: dict
    """
    return read_url(get_uri_from_file_path(file_path))


@lru_cache
def read_resource_file(resource_path: str) -> tuple[dict[str, Any], str]:
    ref = importlib_resources.files("swagger_spec_validator") / resource_path
    with importlib_resources.as_file(ref) as path:
        return read_file(path), path


def read_url(url: str, timeout: float = TIMEOUT_SEC) -> dict[str, Any]:
    with contextlib.closing(urlopen(url, timeout=timeout)) as fh:
        # NOTE: JSON is a subset of YAML so it is safe to read JSON as it is YAML
        return yaml.load(fh.read().decode("utf-8"), Loader=SafeLoader)


class SwaggerValidationError(Exception):
    """Exception raised in case of a validation error."""

    pass


class SwaggerValidationWarning(UserWarning):
    """Warning raised during validation."""

    pass



import glob
import json
import os.path

import jsonschema.exceptions
import pytest

from swagger_spec_validator import SwaggerValidationError
from swagger_spec_validator.validator12 import validate_api_declaration
from swagger_spec_validator.validator12 import validate_resource_listing


def run_json_tests_with_func(json_test_paths, func):
    """Run the specified test function over a list of JSON test files."""

    for json_test_path in sorted(json_test_paths):
        with open(json_test_path) as fd:
            test_data = json.load(fd)

        # Grab last two components from test_path
        # e.g. "api_declarations/array_nested_fail.json"
        test_name = os.sep.join(json_test_path.split(os.sep)[-2:])

        print("Testing %s..." % test_name)

        if test_name.endswith("_pass.json"):
            func(test_data)
        elif test_name.endswith("_fail.json"):
            with pytest.raises(
                (SwaggerValidationError, jsonschema.exceptions.ValidationError)
            ):
                func(test_data)


def test_main():
    run_json_tests_with_func(
        glob.glob("./tests/data/v1.2/api_declarations/*.json"), validate_api_declaration
    )

    run_json_tests_with_func(
        glob.glob("./tests/data/v1.2/resource_listings/*.json"),
        validate_resource_listing,
    )



#
# swagger_spec_validator documentation build configuration file, created by
# sphinx-quickstart on Fri May 13 14:16:02 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os
import sphinx_rtd_theme

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath("../extensions"))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
]
try:
    import sphinx.ext.viewcode

    extensions.append("sphinx.ext.viewcode")
except ImportError:
    pass
try:
    # if it's available, use sphinx_http_domain
    import sphinx_http_domain

    extensions.append("sphinx_http_domain")
except ImportError:
    pass

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Swagger Spec Validator"
copyright = "Yelp Inc"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

base_dir = os.path.join(os.path.dirname(__file__), os.pardir, os.pardir)
about = {}
with open(os.path.join(base_dir, "swagger_spec_validator", "__about__.py")) as f:
    exec(f.read(), about)

version = release = about["__version__"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------
html_theme = "sphinx_rtd_theme"

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


html_static_path = ["_static"]

intersphinx_mapping = {"http://docs.python.org/": None}

html_use_index = True

htmlhelp_basename = "swagger_spec_validatordoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "swagger_spec_validator.tex",
        "swagger_spec_validator Documentation",
        "Yelp Inc",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "index",
        "swagger_spec_validator",
        "Swagger Spec Validator Documentation",
        ["Yelp Inc"],
        1,
    )
]


# Example configuration for intersphinx: refer to the Python standard library.
# intersphinx_mapping = {'http://docs.python.org/': None}



"""
Validate Swagger Specs against the Swagger 1.2 Specification.  The
validator aims to check for full compliance with the Specification.

The validator uses the published jsonschema files for basic structural
validation, augmented with custom validation code where necessary.

https://github.com/swagger-api/swagger-spec/blob/master/versions/1.2.md
"""
from __future__ import annotations

import logging
import os
from typing import Any
from urllib.parse import urlparse

import jsonschema
from jsonschema import RefResolver

from swagger_spec_validator.common import get_uri_from_file_path
from swagger_spec_validator.common import read_resource_file
from swagger_spec_validator.common import read_url
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.common import wrap_exception
from swagger_spec_validator.ref_validators import default_handlers

log = logging.getLogger(__name__)

# Primitives (§4.3.1)
PRIMITIVE_TYPES = ["integer", "number", "string", "boolean"]


def get_model_ids(api_declaration: dict[str, Any]) -> list[str]:
    models = api_declaration.get("models", {})
    return [model["id"] for model in models.values()]


def get_resource_path(url: str, resource: str) -> str:
    """Fetch the complete resource path to get the api declaration.

    :param url: A file or http uri hosting the resource listing.
    :type url: string
    :param resource: Resource path starting with a '/'. eg. '/pet'
    :type resource: string
    :returns: Complete resource path hosting the api declaration.
    """
    if urlparse(url).scheme == "file":
        parent_dir = os.path.dirname(url)

        def resource_file_name(resource: str) -> str:
            assert resource.startswith("/")
            return resource[1:] + ".json"

        path = os.path.join(parent_dir, resource_file_name(resource))
    else:
        path = url + resource

    return path


@wrap_exception
def validate_spec_url(url: str) -> None:
    """Simple utility function to perform recursive validation of a Resource
    Listing and all associated API Declarations.

    This is trivial wrapper function around
    :py:func:`swagger_spec_validator.validate_resource_listing` and
    :py:func:`swagger_spec_validator.validate_api_declaration`.  You are
    encouraged to write your own version of this if required.

    :param url: the URL of the Resource Listing.

    :returns: `None` in case of success, otherwise raises an exception.

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """

    log.info("Validating %s", url)
    validate_spec(read_url(url), url)


def validate_spec(resource_listing: dict[str, Any], url: str) -> None:
    """
    Validates the resource listing, fetches the api declarations and
    consequently validates them as well.

    :type resource_listing: dict
    :param url: url serving the resource listing; needed to resolve api
                declaration path.
    :type url: string

    :returns: `None` in case of success, otherwise raises an exception.

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    validate_resource_listing(resource_listing)

    for api in resource_listing["apis"]:
        path = get_resource_path(url, api["path"])
        log.info("Validating %s", path)
        validate_api_declaration(read_url(path))


def validate_data_type(
    obj: dict[str, Any],
    model_ids: list[str],
    allow_arrays: bool = True,
    allow_voids: bool = False,
    allow_refs: bool = True,
    allow_file: bool = False,
) -> None:
    """Validate an object that contains a data type (§4.3.3).

    Params:
    - obj: the dictionary containing the data type to validate
    - model_ids: a list of model ids
    - allow_arrays: whether an array is permitted in the data type.  This is
      used to prevent nested arrays.
    - allow_voids: whether a void type is permitted.  This is used when
      validating Operation Objects (§5.2.3).
    - allow_refs: whether '$ref's are permitted.  If true, then 'type's
      are not allowed to reference model IDs.
    """

    typ = obj.get("type")
    ref = obj.get("$ref")

    # TODO Use a custom jsonschema.Validator to Validate defaultValue
    # enum, minimum, maximum, uniqueItems
    if typ is not None:
        if typ in PRIMITIVE_TYPES:
            return
        if allow_voids and typ == "void":
            return
        if typ == "array":
            if not allow_arrays:
                raise SwaggerValidationError('"array" not allowed')
            # Items Object (§4.3.4)
            items = obj.get("items")
            if items is None:
                raise SwaggerValidationError('"items" not found')
            validate_data_type(items, model_ids, allow_arrays=False)
            return
        if typ == "File":
            if not allow_file:
                raise SwaggerValidationError(
                    'Type "File" is only valid for form parameters'
                )
            return
        if typ in model_ids:
            if allow_refs:
                raise SwaggerValidationError(
                    'must use "$ref" for referencing "%s"' % typ
                )
            return
        raise SwaggerValidationError('unknown type "%s"' % typ)

    if ref is not None:
        if not allow_refs:
            raise SwaggerValidationError('"$ref" not allowed')
        if ref not in model_ids:
            raise SwaggerValidationError('unknown model id "%s"' % ref)
        return

    raise SwaggerValidationError('no "$ref" or "type" present')


def validate_model(
    model: dict[str, Any], model_name: str, model_ids: list[str]
) -> None:
    """Validate a Model Object (§5.2.7)."""
    # TODO Validate 'sub-types' and 'discriminator' fields
    for required in model.get("required", []):
        if required not in model["properties"]:
            raise SwaggerValidationError(
                'Model "{}": required property "{}" not found'.format(
                    model_name, required
                )
            )

    if model_name != model["id"]:
        error = "model name: {} does not match model id: {}".format(
            model_name, model["id"]
        )
        raise SwaggerValidationError(error)

    for prop_name, prop in model.get("properties", {}).items():
        try:
            validate_data_type(prop, model_ids, allow_refs=True)
        except SwaggerValidationError as e:
            # Add more context to the exception and re-raise
            raise SwaggerValidationError(
                f'Model "{model_name}", property "{prop_name}": {str(e)}'
            )


def validate_parameter(parameter: dict[str, Any], model_ids: list[str]) -> None:
    """Validate a Parameter Object (§5.2.4)."""
    allow_file = parameter.get("paramType") == "form"
    validate_data_type(parameter, model_ids, allow_refs=False, allow_file=allow_file)


def validate_operation(operation: dict[str, Any], model_ids: list[str]) -> None:
    """Validate an Operation Object (§5.2.3)."""
    try:
        validate_data_type(operation, model_ids, allow_refs=False, allow_voids=True)
    except SwaggerValidationError as e:
        raise SwaggerValidationError(
            'Operation "{}": {}'.format(operation["nickname"], str(e))
        )

    for parameter in operation["parameters"]:
        try:
            validate_parameter(parameter, model_ids)
        except SwaggerValidationError as e:
            raise SwaggerValidationError(
                'Operation "%s", parameter "%s": %s'
                % (operation["nickname"], parameter["name"], str(e))
            )


def validate_api(api: dict[str, Any], model_ids: list[str]) -> None:
    """Validate an API Object (§5.2.2)."""
    for operation in api["operations"]:
        validate_operation(operation, model_ids)


def validate_api_declaration(api_declaration: dict[str, Any]) -> None:
    """Validate an API Declaration (§5.2).

    :param api_declaration: a dictionary respresentation of an API Declaration.

    :returns: `None` in case of success, otherwise raises an exception.

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    :raises: :py:class:`jsonschema.exceptions.ValidationError`
    """
    validate_json(api_declaration, "schemas/v1.2/apiDeclaration.json")

    model_ids = get_model_ids(api_declaration)

    for api in api_declaration["apis"]:
        validate_api(api, model_ids)

    for model_name, model in api_declaration.get("models", {}).items():
        validate_model(model, model_name, model_ids)


def validate_resource_listing(resource_listing: dict[str, Any]) -> None:
    """Validate a Resource Listing (§5.1).

    :param resource_listing: a dictionary respresentation of a Resource Listing.

    Note that you will have to invoke `validate_api_declaration` on each
    linked API Declaration.

    :returns: `None` in case of success, otherwise raises an exception.

    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    :raises: :py:class:`jsonschema.exceptions.ValidationError`
    """
    validate_json(resource_listing, "schemas/v1.2/resourceListing.json")


@wrap_exception
def validate_json(json_document: list[Any] | dict[str, Any], schema_path: str) -> None:
    """Validate a json document against a json schema.

    :param json_document: json document in the form of a list or dict.
    :param schema_path: package relative path of the json schema file.
    """
    schema, schema_path = read_resource_file(schema_path)

    resolver = RefResolver(
        base_uri=get_uri_from_file_path(schema_path),
        referrer=schema,
        handlers=default_handlers,
    )
    jsonschema.validate(json_document, schema, resolver=resolver)



import pytest

from swagger_spec_validator.validator12 import get_resource_path


def test_fetch_from_file_uri_success():
    url, resource = "file://bla/bar", "/foo"
    assert "file://bla/foo.json", get_resource_path(url, resource)


def test_fetch_from_http_uri_success():
    url, resource = "http://bla/bar", "/foo"
    assert "http://bla/bar/foo", get_resource_path(url, resource)


def test_fetch_from_bad_file_uri_fail():
    with pytest.raises(AssertionError):
        url, resource = "file://bla/bar", "no_slash"
        get_resource_path(url, resource)



import uuid
from unittest import mock

import importlib_resources

from swagger_spec_validator.common import read_file
from swagger_spec_validator.common import read_resource_file


def test_read_file():
    read_file("./tests/data/v2.0/petstore.json")


def test_read_resource_file(monkeypatch):
    resource_path = str(uuid.uuid4()) + ".json"

    with mock.patch("swagger_spec_validator.common.read_file") as m:
        read_resource_file(resource_path)
        read_resource_file(resource_path)

    m.assert_called_once_with(
        importlib_resources.files("swagger_spec_validator") / resource_path
    )



import json

import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator12 import validate_json


def test_success():
    with open("./tests/data/v1.2/foo/swagger_api.json") as f:
        resource_listing = json.load(f)
    validate_json(resource_listing, "schemas/v1.2/resourceListing.json")

    with open("./tests/data/v1.2/foo/foo.json") as f:
        api_declaration = json.load(f)
    validate_json(api_declaration, "schemas/v1.2/apiDeclaration.json")


def test_failure():
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_json({}, "schemas/v1.2/apiDeclaration.json")
    assert "'swaggerVersion' is a required property" in str(excinfo.value)



from __future__ import annotations

import contextlib
import functools
import logging
from collections.abc import Mapping
from typing import Any
from typing import Callable
from typing import Generator
from typing import TYPE_CHECKING

from jsonschema import validators
from jsonschema.validators import Draft4Validator
from jsonschema.validators import RefResolver

if TYPE_CHECKING:
    from jsonschema.exceptions import _Error
    from jsonschema.validators import _Validator

from swagger_spec_validator import common


log = logging.getLogger(__name__)


default_handlers = {
    "http": common.read_url,
    "https": common.read_url,
    "file": common.read_url,
}


def validate(
    instance: object,
    schema: Mapping[str, Any],
    instance_cls: type[_Validator],
    cls: type[_Validator] | None = None,
    *args: Any,
    **kwargs: Any,
) -> None:
    """This is a carbon-copy of :method:`jsonschema.validate` except that it
    takes two validator classes instead of just one. In the jsonschema
    implementation, `cls` is used to validate both the schema and the
    instance. This changes the behavior to have a separate validator for
    each of schema and instance. Schema should not be validated with the
    custom validator returned by :method:`create_dereffing_validator` because
    it follows $refs.

    :param instance: the instance to validate
    :param schema: the schema to validate with
    :param instance_cls: Validator class to validate instance.
    :param cls: Validator class to validate schema.

    :raises:
        :exc:`ValidationError` if the instance is invalid
        :exc:`SchemaError` if the schema itself is invalid
    """
    if cls is None:
        cls = validators.validator_for(schema)
    cls.check_schema(schema)
    instance_cls(schema, *args, **kwargs).validate(instance)


def create_dereffing_validator(instance_resolver: RefResolver) -> type[_Validator]:
    """Create a customized Draft4Validator that follows $refs in the schema
    being validated (the Swagger spec for a service). This is not to be
    confused with $refs that are in the schema that describes the Swagger 2.0
    specification.

    :param instance_resolver: resolver for the swagger service's spec
    :type instance_resolver: :class:`jsonschema.RefResolver`

    :rtype: Its complicated. See jsonschema.validators.create()
    """
    visited_refs: dict[str, str] = {}

    validators_to_bound = {
        "$ref",
        "additionalProperties",
        "allOf",
        "anyOf",
        "dependencies",
        "maxProperties",
        "minProperties",
        "not",
        "oneOf",
        "patternProperties",
        "properties",
        "required",
        "type",
    }

    bound_validators = {
        k: functools.partial(
            validator_wrapper,
            instance_resolver=instance_resolver,
            visited_refs=visited_refs,
            default_validator_callable=v,
        )
        if k in validators_to_bound
        else v
        for k, v in Draft4Validator.VALIDATORS.items()
    }

    return validators.extend(Draft4Validator, bound_validators)


def validate_schema_value(
    schema: Mapping[str, Any],
    value: Any,
    swagger_resolver: RefResolver | None = None,
) -> None:
    # pass resolver to avoid to refetch schema files
    if swagger_resolver is None:
        swagger_resolver = RefResolver.from_schema(schema)
    create_dereffing_validator(swagger_resolver)(
        schema, resolver=swagger_resolver
    ).validate(value)


@contextlib.contextmanager
def visiting(visited_refs: dict[str, str], ref: str) -> Generator[None, None, None]:
    """Context manager that keeps track of $refs that we've seen during
    validation.

    :param visited_refs: dict of $refs
    :param ref: string $ref value
    """
    visited_refs[ref] = ref
    try:
        yield
    finally:
        del visited_refs[ref]


def validator_wrapper(
    validator: type[_Validator],
    schema_element: Any,
    instance: dict[str, Any],
    schema: Mapping[str, Any],
    instance_resolver: RefResolver,
    visited_refs: dict[str, str],
    default_validator_callable: Callable,
) -> Generator[_Error, None, None]:
    """Generator function that parameterizes default_validator_callable.

    :type validator: :class:`jsonschema.validators.Validator`
    :param schema_element: The schema element that is passed in to each
        specific validator callable aka the 2nd arg in each
        jsonschema._validators.* callable.
    :param instance: The fragment of the swagger service spec that is being
        validated.
    :param schema: The fragment of the swagger jsonschema spec that describes
        is used for validation.
    :param instance_resolver: Resolves refs in the swagger service spec
    :param visited_refs: Keeps track of visisted refs during validation of
        the swagger service spec.
    :param default_validator_callable: jsonschema._validators.* callable
    """
    yield from deref_and_validate(
        validator,
        schema_element,
        instance,
        schema,
        instance_resolver,
        visited_refs,
        default_validator_callable,
    )


def deref_and_validate(
    validator: type[_Validator],
    schema_element: Any,
    instance: dict[str, Any],
    schema: Mapping[str, Any],
    instance_resolver: RefResolver,
    visited_refs: dict[str, str],
    default_validator_callable: Callable,
) -> Generator[_Error, None, None]:
    """Generator function that dereferences instance if it is a $ref before
    passing it downstream for actual validation. When a cyclic ref is detected,
    short-circuit and return.

    :type validator: :class:`jsonschema.validators.Validator`
    :param schema_element: The schema element that is passed in to each
        specific validator callable aka the 2nd arg in each
        jsonschema._validators.* callable.
    :param instance: The fragment of the swagger service spec that is being
        validated.
    :param schema: The fragment of the swagger jsonschema spec that describes
        is used for validation.
    :param instance_resolver: Resolves refs in the swagger service spec
    :param visited_refs: Keeps track of visisted refs during validation of
        the swagger service spec.
    :param default_validator_callable: jsonschema._validators.* callable
    """
    if (
        isinstance(instance, dict)
        and "$ref" in instance
        and isinstance(instance["$ref"], str)
    ):
        ref = instance["$ref"]
        # Annotate $ref dict with scope - used by custom validations
        # We still need to attach the scope even if this is a cycle, as otherwise there are cases
        # with specs split into multiple files where it can't be dereferenced properly
        attach_scope(instance, instance_resolver)

        if ref in visited_refs:
            log.debug("Found cycle in %s", ref)
            return

        with visiting(visited_refs, ref):
            with instance_resolver.resolving(ref) as target:
                yield from default_validator_callable(
                    validator, schema_element, target, schema
                )

    else:
        yield from default_validator_callable(
            validator, schema_element, instance, schema
        )


def attach_scope(ref_dict: dict[str, Any], instance_resolver: RefResolver) -> None:
    """Attach scope to each $ref we encounter so that the $ref can be
    resolved by custom validations done outside the scope of jsonscema
    validations.

    :param ref_dict: dict with $ref key
    :type instance_resolver: :class:`jsonschema.RefResolver`
    """
    if "x-scope" in ref_dict:
        log.debug("Ref %s already has scope attached", ref_dict["$ref"])
        return
    log.debug("Attaching x-scope to %s", ref_dict)
    ref_dict["x-scope"] = list(instance_resolver._scopes_stack)  # type: ignore


@contextlib.contextmanager
def in_scope(
    resolver: RefResolver, ref_dict: dict[str, Any]
) -> Generator[None, None, None]:
    """Context manager to assume the given scope for the passed in resolver.

    The resolver's original scope is restored when exiting the context manager.

    :type resolver: :class:`jsonschema.RefResolver
    :type ref_dict: dict
    """
    if "x-scope" not in ref_dict:
        yield
    else:
        saved_scope_stack = resolver._scopes_stack  # type: ignore
        try:
            resolver._scopes_stack = ref_dict["x-scope"]  # type: ignore
            yield
        finally:
            resolver._scopes_stack = saved_scope_stack  # type: ignore



import pytest

from swagger_spec_validator import validator12
from swagger_spec_validator import validator20
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.util import get_validator


def test_version_missing():
    with pytest.raises(SwaggerValidationError) as excinfo:
        get_validator({}, "http://foo.com")
    assert "missing" in str(excinfo.value)


def test_1_dot_x_version_not_supported():
    with pytest.raises(SwaggerValidationError) as excinfo:
        get_validator({"swaggerVersion": "0.9"}, "http://foo.com")
    assert "not supported" in str(excinfo.value)


def test_2_dot_x_version_not_supported():
    with pytest.raises(SwaggerValidationError) as excinfo:
        get_validator({"swagger": "1.2"}, "http://foo.com")
    assert "not supported" in str(excinfo.value)


def test_both_swagger_1_dot_x_and_2_dot_x_version_keys_found():
    with pytest.raises(SwaggerValidationError) as excinfo:
        spec = {"swagger": "2.0", "swaggerVersion": "1.2"}
        get_validator(spec, "http://foo.com")
    assert "not both" in str(excinfo.value)


def test_success_12():
    spec = {"swaggerVersion": "1.2"}
    assert validator12 == get_validator(spec, "http://foo.com") == validator12


def test_success_20():
    spec = {"swagger": "2.0"}
    assert validator20 == get_validator(spec, "http://foo.com")



import os

from setuptools import find_packages, setup

base_dir = os.path.dirname(__file__)

about = {}
with open(os.path.join(base_dir, "swagger_spec_validator", "__about__.py")) as f:
    exec(f.read(), about)


install_requires = [
    "jsonschema",
    "pyyaml",
    "typing-extensions",
    "importlib-resources >= 1.3",
]


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__summary__"],
    url=about["__uri__"],
    author=about["__author__"],
    author_email=about["__email__"],
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={
        "swagger_spec_validator": [
            "swagger_spec_validator/schemas/v1.2/*",
            "swagger_spec_validator/schemas/v2.0/*",
            "swagger_spec_validator/py.typed",
        ],
    },
    python_requires=">=3.8",
    include_package_data=True,
    install_requires=install_requires,
    license=about["__license__"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import pytest

from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator20 import validate_parameters


@pytest.mark.parametrize(
    "top_level_parameters",
    [
        {},
        {
            "Offset": {
                "in": "query",
                "name": "offset",
                "type": "integer",
            },
        },
        {
            "AcceptedLanguage": {
                "in": "header",
                "name": "Accepted-Language",
                "type": "string",
            },
        },
    ],
)
def test_valid_top_level_parameters(top_level_parameters):
    validate_parameters(top_level_parameters, deref=lambda x: x)


@pytest.mark.parametrize(
    "top_level_parameters, expected_exception_string",
    [
        [
            {
                "Offset": {
                    "in": "body",
                    "name": "body",
                },
            },
            "Body parameter in `#/parameters/Offset` does not specify `schema`.",
        ],
        [
            {
                "ParameterWithoutType": {
                    "in": "header",
                    "name": "a-random-parameter",
                },
            },
            "Non-Body parameter in `#/parameters/ParameterWithoutType` does not specify `type`.",
        ],
        [
            {
                "ListOfIntegers": {
                    "in": "header",
                    "name": "Accepted-Language",
                    "type": "array",
                },
            },
            "Non-Body array parameter in `#/parameters/ListOfIntegers` does not specify `items`.",
        ],
    ],
)
def test_invalid_top_level_parameters(top_level_parameters, expected_exception_string):
    with pytest.raises(SwaggerValidationError) as excinfo:
        validate_parameters(top_level_parameters, deref=lambda x: x)

    assert str(excinfo.value) == expected_exception_string



from unittest import mock

import pytest

from .validate_spec_url_test import make_mock_responses
from .validate_spec_url_test import read_contents
from swagger_spec_validator.common import get_uri_from_file_path
from swagger_spec_validator.common import SwaggerValidationError
from swagger_spec_validator.validator12 import validate_data_type
from swagger_spec_validator.validator12 import validate_model
from swagger_spec_validator.validator12 import validate_parameter
from swagger_spec_validator.validator12 import validate_spec
from tests import TESTS_BASE_PATH


RESOURCE_LISTING_FILE = TESTS_BASE_PATH + "/data/v1.2/foo/swagger_api.json"
API_DECLARATION_FILE = TESTS_BASE_PATH + "/data/v1.2/foo/foo.json"


def get_resource_listing():
    return read_contents(RESOURCE_LISTING_FILE)


def test_http_success():
    mock_responses = make_mock_responses([API_DECLARATION_FILE])

    with mock.patch(
        "swagger_spec_validator.validator12.read_url", side_effect=mock_responses
    ) as mock_read_url:
        validate_spec(get_resource_listing(), "http://localhost/api-docs")
        mock_read_url.assert_called_once_with("http://localhost/api-docs/foo")


def test_file_uri_success():
    mock_string = "swagger_spec_validator.validator12.validate_api_declaration"
    with mock.patch(mock_string) as mock_api:
        validate_spec(
            get_resource_listing(),
            get_uri_from_file_path(RESOURCE_LISTING_FILE),
        )

        expected = read_contents(API_DECLARATION_FILE)
        mock_api.assert_called_once_with(expected)


def test_validate_parameter_type_file_in_form():
    parameter = {
        "paramType": "form",
        "name": "what",
        "type": "File",
    }
    # lack of errors is success
    validate_parameter(parameter, [])


def test_validate_parameter_type_file_in_body():
    parameter = {
        "paramType": "body",
        "name": "what",
        "type": "File",
    }
    with pytest.raises(
        SwaggerValidationError, match='Type "File" is only valid for form parameters'
    ):
        validate_parameter(parameter, [])


def test_validate_data_type_is_model():
    model_id = "MyModelId"
    model_ids = [model_id, "OtherModelId"]
    obj = {"type": model_id}
    # lack of error is success
    validate_data_type(obj, model_ids, allow_refs=False)


def test_validate_model_matches_id():
    model = {"id": "mysupermodel"}
    model_name = "mymodel"
    model_ids = ""

    with pytest.raises(
        SwaggerValidationError,
        match="model name: mymodel does not match model id: mysupermodel",
    ):
        validate_model(model, model_name, model_ids)



import pytest

from swagger_spec_validator.common import SwaggerValidationWarning
from swagger_spec_validator.validator20 import validate_references


@pytest.mark.parametrize(
    "raw_spec",
    [
        {"$ref": "#/"},
        {"description": "Description sibling is acceptable", "$ref": "#/"},
        [{"$ref": "#/"}],
    ],
)
def test_validate_valid_references(recwarn, raw_spec):
    validate_references(raw_spec=raw_spec, deref=lambda x: x)
    assert len(recwarn) == 0


@pytest.mark.parametrize(
    "raw_spec, expected_warning_messages",
    [
        (
            {"sibling-attribute": "", "$ref": "#/"},
            (
                'Found "$ref: #/" with siblings that will be overwritten. '
                "See https://stackoverflow.com/a/48114924 for more information. (path #)",
            ),
        ),
        (
            [{"sibling-attribute": "", "$ref": "#/"}],
            (
                'Found "$ref: #/" with siblings that will be overwritten. '
                "See https://stackoverflow.com/a/48114924 for more information. (path #/0)",
            ),
        ),
        (
            {"$ref": None},
            (
                "Identified $ref with None value. This is usually an error, although technically it might be allowed. "
                "(path: #)",
            ),
        ),
        (
            {"key": {"$ref": None}},
            (
                "Identified $ref with None value. This is usually an error, although technically it might be allowed. "
                "(path: #/key)",
            ),
        ),
        (
            {"key": [{"sibling-attribute": 1, "$ref": None}]},
            (
                'Found "$ref: None" with siblings that will be overwritten. '
                "See https://stackoverflow.com/a/48114924 for more information. (path #/key/0)",
                "Identified $ref with None value. This is usually an error, although technically it might be allowed. "
                "(path: #/key/0)",
            ),
        ),
    ],
)
def test_validate_references_to_warn(raw_spec, expected_warning_messages):
    with pytest.warns(SwaggerValidationWarning) as warninfo:
        validate_references(raw_spec=raw_spec, deref=lambda x: x)

    assert sorted(expected_warning_messages) == sorted(
        str(warning.message) for warning in warninfo.list
    )
